/* $Id: virtual.c,v 1.4 2002/10/31 21:36:50 mikpe Exp $
 * Library interface to Linux x86 Performance-Monitoring Counters.
 * Virtual per-process performance counters.
 *
 * Copyright (C) 1999-2002  Mikael Pettersson
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>
#include "libperfctr.h"

#define PAGE_SIZE	4096

static inline int sys_vperfctr_info(int fd, struct perfctr_cpu_info *info)
{
    /* pass vperfctr's fd to allow emulation for perfctr-2.4 */
    return sys_vperfctr(VPERFCTR_INFO, fd, info);
}

static inline int sys_vperfctr_sample(int fd)
{
    return sys_vperfctr(VPERFCTR_SAMPLE, fd, NULL);
}

static inline int sys_vperfctr_control(int fd, struct vperfctr_control *control)
{
    return sys_vperfctr(VPERFCTR_CONTROL, fd, control);
}

static inline int sys_vperfctr_iresume(int fd)
{
    return sys_vperfctr(VPERFCTR_IRESUME, fd, NULL);
}

static inline int sys_vperfctr_unlink(int fd)
{
    return sys_vperfctr(VPERFCTR_UNLINK, fd, NULL);
}

static inline int sys_vperfctr_open_self(void)
{
    return sys_vperfctr(VPERFCTR_OPEN, 0, NULL);
}

static inline int sys_vperfctr_creat_self(void)
{
    return sys_vperfctr(VPERFCTR_CREAT, 0, NULL);
}

/*
 * Operations on the process' own virtual-mode perfctrs.
 */

struct vperfctr {
    /* XXX: point to &vperfctr_state.cpu_state instead? */
    volatile const struct vperfctr_state *kstate;
    int fd;
    unsigned char have_rdpmc;
};

struct vperfctr *vperfctr_open(void)
{
    struct vperfctr *perfctr;
    struct perfctr_cpu_info info;
    int isnew;

    perfctr = malloc(sizeof(*perfctr));
    if( !perfctr )
	return NULL;
    isnew = 1;
    perfctr->fd = sys_vperfctr_open_self();
    if( perfctr->fd < 0 ) {
	isnew = 0;
	perfctr->fd = sys_vperfctr_creat_self();
	if( perfctr->fd < 0 )
	    goto err_perfctr;
    }
    if( sys_vperfctr_info(perfctr->fd, &info) != 0 )
	goto err_fd;
    perfctr->have_rdpmc = (info.features & PERFCTR_FEATURE_RDPMC) != 0;
    perfctr->kstate = mmap(NULL, PAGE_SIZE, PROT_READ,
			   MAP_SHARED, perfctr->fd, 0);
    if( perfctr->kstate != MAP_FAILED ) {
	if( perfctr->kstate->magic == VPERFCTR_MAGIC )
	    return perfctr;
	fprintf(stderr, __FILE__ ":%s: kstate version mismatch, kernel %#x, expected %#x\n",
		__FUNCTION__, perfctr->kstate->magic, VPERFCTR_MAGIC);
	munmap((void*)perfctr->kstate, PAGE_SIZE);
    }
 err_fd:
    if( isnew )
	vperfctr_unlink(perfctr);
    close(perfctr->fd);
 err_perfctr:
    free(perfctr);
    return NULL;
}

int vperfctr_info(const struct vperfctr *vperfctr, struct perfctr_cpu_info *info)
{
    return sys_vperfctr_info(vperfctr->fd, info);
}

#define rdtscl(low)	\
	__asm__ __volatile__("rdtsc" : "=a"(low) : : "edx")
#define rdpmcl(ctr,low)	\
	__asm__ __volatile__("rdpmc" : "=a"(low) : "c"(ctr) : "edx")

#if (__GNUC__ < 2) ||  (__GNUC__ == 2 && __GNUC_MINOR__ < 96)
#define __builtin_expect(x, expected_value) (x)
#endif

#define likely(x)	__builtin_expect((x),1)
#define unlikely(x)	__builtin_expect((x),0)

unsigned long long vperfctr_read_tsc(const struct vperfctr *self)
{
    unsigned long long sum;
    unsigned int tsc0, tsc1, now;
    volatile const struct vperfctr_state *kstate;

    kstate = self->kstate;
    if( likely(kstate->cpu_state.cstatus != 0) ) {
	tsc0 = kstate->cpu_state.start.tsc;
    retry:
	rdtscl(now);
	sum = kstate->cpu_state.sum.tsc;
	tsc1 = kstate->cpu_state.start.tsc;
	if( likely(tsc1 == tsc0) )
	    return sum += (now - tsc0);
	tsc0 = tsc1;
	goto retry; /* better gcc code than with a do{}while() loop */
    }
    return kstate->cpu_state.sum.tsc;
}

unsigned long long vperfctr_read_pmc(const struct vperfctr *self, unsigned int i)
{
    unsigned long long sum;
    unsigned int start, now;
    unsigned int tsc0, tsc1;
    volatile const struct vperfctr_state *kstate;
    unsigned int cstatus;

    kstate = self->kstate;
    cstatus = kstate->cpu_state.cstatus;
    /* gcc 3.0 generates crap code for likely(E1 && E2) :-( */
    if( perfctr_cstatus_has_tsc(cstatus) && self->have_rdpmc ) {
	 tsc0 = kstate->cpu_state.start.tsc;
    retry:
	 rdpmcl(kstate->cpu_state.control.pmc_map[i], now);
	 start = kstate->cpu_state.start.pmc[i];
	 sum = kstate->cpu_state.sum.pmc[i];
	 tsc1 = kstate->cpu_state.start.tsc;
	 if( likely(tsc1 == tsc0) ) {
	      return sum += (now - start);
	 }
	 tsc0 = tsc1;
	 goto retry;
    }
    if( cstatus != 0 )
	sys_vperfctr_sample(self->fd);
    return kstate->cpu_state.sum.pmc[i];
}

void vperfctr_read_ctrs(const struct vperfctr *self,
			struct perfctr_sum_ctrs *sum)
{
    unsigned int tsc0, tsc1, now;
    unsigned int cstatus, nrctrs;
    volatile const struct vperfctr_state *kstate;
    int i;

    /* Fast path is impossible if the TSC isn't being sampled (bad idea,
       but on WinChip you don't have a choice), or at least one PMC is
       enabled but the CPU doesn't have RDPMC. */
    kstate = self->kstate;
    cstatus = kstate->cpu_state.cstatus;
    nrctrs = perfctr_cstatus_nrctrs(cstatus);
    if( perfctr_cstatus_has_tsc(cstatus) && (!nrctrs || self->have_rdpmc) ) {
    retry:
	tsc0 = kstate->cpu_state.start.tsc;
	rdtscl(now);
	sum->tsc = kstate->cpu_state.sum.tsc + (now - tsc0);
	for(i = nrctrs; --i >= 0;) {
	    rdpmcl(kstate->cpu_state.control.pmc_map[i], now);
	    sum->pmc[i] = kstate->cpu_state.sum.pmc[i] + (now - kstate->cpu_state.start.pmc[i]);
	}
	if( likely(tsc0 == kstate->cpu_state.start.tsc) )
	    return;
	goto retry;
    }
    sys_vperfctr_sample(self->fd);
    tsc1 = kstate->cpu_state.start.tsc;
    do {
	tsc0 = tsc1;
	sum->tsc = kstate->cpu_state.sum.tsc;
	for(i = 0; i < nrctrs; ++i)
	    sum->pmc[i] = kstate->cpu_state.sum.pmc[i];
	tsc1 = kstate->cpu_state.start.tsc;
    } while( tsc1 != tsc0 );
}

int vperfctr_read_state(const struct vperfctr *self, struct perfctr_sum_ctrs *sum,
			struct vperfctr_control *control)
{
    unsigned int prev_tsc, next_tsc;
    volatile const struct vperfctr_state *kstate;

    sys_vperfctr_sample(self->fd);
    kstate = self->kstate;
    next_tsc = kstate->cpu_state.start.tsc;
    do {
	prev_tsc = next_tsc;
	/* XXX: this copies more than necessary */
	if( sum )
	    *sum = kstate->cpu_state.sum;
	if( control ) {
	    control->si_signo = kstate->si_signo;
	    control->cpu_control = kstate->cpu_state.control;
	}
	next_tsc = kstate->cpu_state.start.tsc;
    } while( next_tsc != prev_tsc );
    return 0;
}

int vperfctr_control(const struct vperfctr *perfctr,
		     struct vperfctr_control *control)
{
    return sys_vperfctr_control(perfctr->fd, control);
}

int vperfctr_stop(const struct vperfctr *perfctr)
{
    struct vperfctr_control control;
    memset(&control, 0, sizeof control);
    return vperfctr_control(perfctr, &control);
}

int vperfctr_is_running(const struct vperfctr *perfctr)
{
    return perfctr->kstate->cpu_state.cstatus != 0;
}

int vperfctr_iresume(const struct vperfctr *perfctr)
{
    return sys_vperfctr_iresume(perfctr->fd);
}

int vperfctr_unlink(const struct vperfctr *perfctr)
{
    return sys_vperfctr_unlink(perfctr->fd);
}

void vperfctr_close(struct vperfctr *perfctr)
{
    munmap((void*)perfctr->kstate, PAGE_SIZE);
    close(perfctr->fd);
    free(perfctr);
}
